/*
** Copyright (C) 1999 Christian Werner
**
** This program and library is free software; you can redistribute it and/or
** modify it under the terms of the GNU (Library) General Public License
** as published by the Free Software Foundation; either version 2
** of the License, or any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
** GNU (Library) General Public License for more details.
**
** You should have received a copy of the GNU (Library) General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

package swishej;

import java.io.*;
import java.net.*;
import java.util.zip.*;

class GZBuf {
    /**
     * Expanded size of data.
     */
    int size;

    /**
     * Expanded data or null.
     */
    byte buffer[];

    /**
     * Collapsed data or null.
     */
    byte gzbuf[];

    /**
     * Make buffer given expanded size.
     * @param size expanded size
     */
    GZBuf(int size) {
        this.size = size;
        buffer = new byte[size];
	gzbuf = null;
    }

    /**
     * Collapse buffer if not already collapsed.
     * @exception IOException if an error occurred
     */
    void collapse() throws IOException {
        if (buffer == null) {
	    return;
	}
	ByteArrayOutputStream bs = new ByteArrayOutputStream();
	GZIPOutputStream gz = null;
	try {
	    gz = new GZIPOutputStream(bs);
	} catch (IOException e) {
	    return;
	}
	try {
	    gz.write(buffer, 0, buffer.length);
	    gz.finish();
	} catch (IOException e) {
	    return;
	}
	gzbuf = bs.toByteArray();
	gz.close();
	buffer = null;
    }

    /**
     * Expand buffer if not already expanded.
     * @exception IOException if an error occurred
     */
    void expand() throws IOException {
        if (buffer != null) {
	    return;
	}
	ByteArrayInputStream bs = new ByteArrayInputStream(gzbuf);
	GZIPInputStream gz = new GZIPInputStream(bs);
	buffer = new byte[size];
	int got = 0;
	while (got < size) {
	    int nread = gz.read(buffer, got, size - got);
	    got += nread;
	}
	gz.close();
	gzbuf = null;
    }
}

/**
 * Reader for reading a SWISH-E index file from a byte array which
 * gets filled from a URL.
 * @see swishej.Reader
 */

class UReader implements Reader {
    /**
     * Default expanded chunk size.
     */
    static final int CHUNKSIZE = 64 * 1024;

    /**
     * File data buffer.
     */
    byte buffer[];

    /**
     * Chunks for managing data.
     */
    GZBuf gzbufs[];

    /**
     * Total data size
     */
    int total;

    /**
     * Current file position.
     */
    long pos;

    /**
     * Create an index file reader from given URL. The content of that
     * URL is read into internal buffers, which can later be operated on
     * by the read, seek, and tell  methods.
     * @param url URL to fill the buffers from.
     * @exception MalformedURLException if the URL is invalid
     * @exception IOException if a read error occurred
     */
    UReader(URL url) throws MalformedURLException, IOException {
	gzbufs = null;
	pos = 0;
        URLConnection uc = url.openConnection();
	total = uc.getContentLength();
	int nbufs = (total + CHUNKSIZE - 1) / CHUNKSIZE;
	gzbufs = new GZBuf[nbufs];
	int got = 0;
	while (got != total) {
	    int max = total - got;
	    if (max > CHUNKSIZE) {
	        max = CHUNKSIZE;
	    }
	    int i = got / CHUNKSIZE;
	    if (gzbufs[i] == null) {
		gzbufs[i] = new GZBuf(max);
	    }
	    int bpos = got - i * CHUNKSIZE;
	    max = gzbufs[i].buffer.length - bpos;
	    int nread = uc.getInputStream().read(gzbufs[i].buffer,
						 bpos, max);
	    if (nread == -1) {
	        throw new IOException("premature eof at " + got);
	    }
	    got += nread;
	    if (i != 0 && got == i * CHUNKSIZE + gzbufs[i].buffer.length) {
	        gzbufs[i].collapse();
	    }
	}
    }

    /**
     * Read an array of bytes.
     * @param b byte array to be filled by the read
     * @return number of bytes read
     * @exception IOException if an error occurred
     */
    public int read(byte[] b) throws IOException {
        long max = b.length;
	long nread = 0;
	while (pos < total && nread < max) {
	    int oldindex = (int) pos / CHUNKSIZE;
	    int bpos = (int) pos - oldindex * CHUNKSIZE;
	    gzbufs[oldindex].expand();
	    int bmax = (int) (max - nread);
	    if (gzbufs[oldindex].buffer.length - bpos < bmax) {
	        bmax = gzbufs[oldindex].buffer.length - bpos;
	    }
	    System.arraycopy(gzbufs[oldindex].buffer, bpos, b,
			     (int) nread, (int) bmax);
	    pos += bmax;
	    nread += bmax;
	    int newindex = (int) pos / CHUNKSIZE;
	    if (newindex != oldindex) {
		gzbufs[oldindex].collapse();
		gzbufs[newindex].expand();
	    }
	}
	return (int) nread; 
    }

    /**
     * Seek to file position.
     * @param pos long position measured from begin of file, must be positive
     * @exception IOException if an error occurred
     */
    public void seek(long pos) throws IOException {
        if (pos < 0 || pos >= total) {
	    throw new IOException();
	}
	int oldindex = (int) this.pos / CHUNKSIZE;
	int newindex = (int) pos / CHUNKSIZE;
	if (newindex != oldindex) {
	    gzbufs[oldindex].collapse();
	    gzbufs[newindex].expand();
	}
	this.pos = pos;
    }

    /**
     * Retrieve current file position
     * @return current file position
     */
    public long tell() {
        return pos;
    }
}
